/*
 * Copyright (c) 2009, Stefan Krah
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <iostream>
#include <sstream> 
#include "apfloat.h"
#include "apint.h"
#include "mpdecimal.h"


using namespace std;


static char *
randstr(size_t prec)
{
	char *cp;
	size_t i;

	if ((cp = (char *)malloc(prec+1)) == NULL) {
		fprintf(stderr, "out of memory\n");
		exit(1);
	}

	for (i = 0; i < prec; i++) {
		cp[i] = random()%10 + '0';
	}
	cp[i] = '\0';

	return cp;
}

int
main(void)
{
	mpd_context_t ctx;
	mpd_t *a, *b, *c;
	apfloat x, y, z;
	apint ix, iy, iz;
	long small_prec[] = {38, LONG_MAX};
	long huge_prec[] = {5000, 100000, 1000000, 10000000, 100000000, LONG_MAX};
	long prec;
	clock_t start_clock, end_clock;
	char *s, *mpdresult;
	ostringstream o;
	int i, j;

	mpd_init(&ctx, MPD_MINALLOC_MAX);
	mpd_maxcontext(&ctx);

	a = mpd_new(&ctx);
	b = mpd_new(&ctx);
	c = mpd_new(&ctx);


	printf("# ======================================================================\n");
	printf("#                              mul(x, y)\n");
	printf("# ======================================================================\n\n");

	i = 0;
	while (1) {

		prec = huge_prec[i++];
		if (prec == LONG_MAX) {
			break;
		}
		printf("########## prec: %ld  iter: 1\n\n", prec);

		ctx.prec = 2*prec;

		s = randstr(prec);
		mpd_set_string(a, s, &ctx);
        	ix = apint(s);
		free(s);

		s = randstr(prec);
		mpd_set_string(b, s, &ctx);
        	iy = apint(s);
		free(s);

		start_clock = clock();
		mpd_mul(c, a, b, &ctx);
		end_clock = clock();
		printf("mpd calc time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);
		start_clock = clock();
		mpdresult = mpd_to_sci(c, 1);
		end_clock = clock();
		printf("mpd conv time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

		start_clock = clock();
		iz = ix * iy;
		end_clock = clock();
	        printf("apfloat calc time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);
		start_clock = clock();
		o << iz << ends;
		end_clock = clock();
	        printf("apfloat conv time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

		if (strcmp(mpdresult, o.str().c_str()) != 0) {
			fprintf(stderr, "fail in multiply\n");
			exit(1);
		}

		free(mpdresult);
		o.str("");
	}

	printf("# ======================================================================\n");
	printf("#                              div(x, y)\n");
	printf("# ======================================================================\n\n");

	i = 0;
	while (1) {

		prec = huge_prec[i++];
		if (prec == LONG_MAX) {
			break;
		}
		printf("########## prec: %ld  iter: 1\n\n", prec);

		ctx.prec = prec;

		s = randstr(prec);
		mpd_set_string(a, s, &ctx);
        	x = apfloat(s, prec);
		free(s);

		s = randstr(prec/4);
		mpd_set_string(b, s, &ctx);
        	y = apfloat(s, prec);
		free(s);

		start_clock = clock();
		mpd_div(c, a, b, &ctx);
		end_clock = clock();
		printf("mpd time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

		start_clock = clock();
		z = x / y;
		end_clock = clock();
	        printf("apfloat time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);
	}


	printf("# ======================================================================\n");
	printf("#                              sqrt(x)\n");
	printf("# ======================================================================\n\n");

	i = 0;
	while (1) {

		prec = small_prec[i++];
		if (prec == LONG_MAX) {
			break;
		}
		printf("########## prec: %ld  iter: 100000\n\n", prec);

		ctx.prec = prec;

		s = randstr(prec);
		mpd_set_string(a, s, &ctx);
        	x = apfloat(s, prec);
		free(s);

		start_clock = clock();
		for (j = 0; j < 100000; j++) {
			mpd_sqrt(b, a, &ctx);
		}
		end_clock = clock();

		mpd_print(b);
		printf("mpd time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

		start_clock = clock();
		for (j = 0; j < 100000; j++) {
			y = sqrt(x);
		}
		end_clock = clock();

		cout << pretty << y << endl;
		printf("apfloat time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);
	}


	printf("########## prec: 1000000  iter: 1\n\n");

	prec = 1000000;
	ctx.prec = prec;

	s = randstr(prec);
	mpd_set_string(a, s, &ctx);
       	x = apfloat(s, prec);
	free(s);

	start_clock = clock();
	mpd_sqrt(b, a, &ctx);
	end_clock = clock();
	printf("mpd time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

	start_clock = clock();
	y = sqrt(x);
	end_clock = clock();
	printf("apfloat time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);


	printf("########## prec: 10000000  iter: 1\n\n");

	prec = 10000000;
	ctx.prec = prec;

	s = randstr(prec);
	mpd_set_string(a, s, &ctx);
       	x = apfloat(s, prec);
	free(s);

	start_clock = clock();
	mpd_sqrt(b, a, &ctx);
	end_clock = clock();
	printf("mpd time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

	start_clock = clock();
	y = sqrt(x);
	end_clock = clock();
	printf("apfloat time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);


	printf("# ======================================================================\n");
	printf("#                              invroot(x)\n");
	printf("# ======================================================================\n\n");

	printf("########## prec: 1000000  iter: 1\n\n");

	prec = 1000000;
	ctx.prec = prec;

	s = randstr(prec);
	mpd_set_string(a, s, &ctx);
       	x = apfloat(s, prec);
	free(s);

	start_clock = clock();
	mpd_invroot(b, a, &ctx);
	end_clock = clock();
	printf("mpd time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

	start_clock = clock();
	y = invroot(x, 2);
	end_clock = clock();
	printf("apfloat time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

	printf("########## prec: 10000000  iter: 1\n\n");

	prec = 10000000;
	ctx.prec = prec;

	s = randstr(prec);
	mpd_set_string(a, s, &ctx);
       	x = apfloat(s, prec);
	free(s);

	start_clock = clock();
	mpd_invroot(b, a, &ctx);
	end_clock = clock();
	printf("mpd time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

	start_clock = clock();
	y = invroot(x, 2);
	end_clock = clock();
	printf("apfloat time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);


	printf("# ======================================================================\n");
	printf("#                              exp(x)\n");
	printf("# ======================================================================\n\n");

	i = 0;
	while (1) {

		prec = small_prec[i++];
		if (prec == LONG_MAX) {
			break;
		}
		printf("########## prec: %ld  iter: 100000\n\n", prec);

		ctx.prec = prec;

		s = randstr(8);
		mpd_set_string(a, s, &ctx);
        	x = apfloat(s, prec);
		free(s);

		start_clock = clock();
		for (j = 0; j < 100000; j++) {
			mpd_exp(b, a, &ctx);
		}
		end_clock = clock();

		mpd_print(b);
		printf("mpd time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

		start_clock = clock();
		for (j = 0; j < 100000; j++) {
			y = exp(x);
		}
		end_clock = clock();

		cout << y << endl;
		printf("apfloat time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);
	}


	printf("########## prec: 5000  iter: 1\n\n");

	prec = 5000;
	ctx.prec = prec;

	s = randstr(8);
	mpd_set_string(a, s, &ctx);
       	x = apfloat(s, prec);
	free(s);

	start_clock = clock();
	mpd_exp(b, a, &ctx);
	end_clock = clock();
	printf("mpd time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

	start_clock = clock();
	y = exp(x);
	end_clock = clock();
	printf("apfloat time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);


	printf("# ======================================================================\n");
	printf("#                              ln(x)\n");
	printf("# ======================================================================\n\n");

	i = 0;
	while (1) {

		prec = small_prec[i++];
		if (prec == LONG_MAX) {
			break;
		}
		printf("########## prec: %ld  iter: 100000\n\n", prec);

		ctx.prec = prec;

		s = randstr(prec);
		mpd_set_string(a, s, &ctx);
        	x = apfloat(s, prec);
		free(s);

		start_clock = clock();
		for (j = 0; j < 100000; j++) {
			mpd_ln(b, a, &ctx);
		}
		end_clock = clock();

		mpd_print(b);
		printf("mpd time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

		start_clock = clock();
		for (j = 0; j < 100000; j++) {
			y = log(x);
		}
		end_clock = clock();

		cout << pretty << y << endl;
		printf("apfloat time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);
	}


	printf("########## prec: 5000  iter: 1\n\n");

	prec = 5000;
	ctx.prec = prec;

	s = randstr(prec);
	mpd_set_string(a, s, &ctx);
       	x = apfloat(s, prec);
	free(s);

	start_clock = clock();
	mpd_ln(b, a, &ctx);
	end_clock = clock();
	printf("mpd time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

	start_clock = clock();
	y = log(x);
	end_clock = clock();
	printf("apfloat time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);


	return 0;
}



