/*
 * Copyright (c) 2009, Stefan Krah
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "mpdecimal.h"


/* Nonsense version of escape-time algorithm for calculating a mandelbrot
 * set. Just for benchmarking. */
void
color_point(mpd_t *x0, mpd_t *y0, int maxiter, mpd_context_t *ctx)
{
	mpd_t *x, *y, *sq_x, *sq_y, *two, *four;
	int i;

	two = mpd_new(ctx);
	four = mpd_new(ctx);
	mpd_set_u32(two, 2, ctx);
	mpd_set_u32(four, 4, ctx);

	x = mpd_new(ctx);
	y = mpd_new(ctx);
	mpd_copy(x, x0, ctx);
	mpd_copy(y, y0, ctx);

	sq_x = mpd_new(ctx);
	sq_y = mpd_new(ctx);
	mpd_mul(sq_x, x, x, ctx);
	mpd_mul(sq_y, y, y, ctx);

	for (i = 0; i < maxiter; i++) {
		mpd_mul(y, x, y, ctx);
		mpd_mul(y, y, two, ctx);
		mpd_add(y, y, y0, ctx);
		mpd_sub(x, sq_x, sq_y, ctx);
		mpd_add(x, x, x0, ctx);
		mpd_mul(sq_x, x, x, ctx);
		mpd_mul(sq_y, y, y, ctx);
	}

	mpd_copy(x0, x, ctx);

	mpd_del(x);
	mpd_del(y);
	mpd_del(sq_x);
	mpd_del(sq_y);
	mpd_del(two);
	mpd_del(four);
}

int
main(int argc, char **argv)
{
	mpd_context_t ctx;
	mpd_t *x0, *y0;
	uint32_t prec = 19;
	long iter = 10000000;
	clock_t start_clock, end_clock;

	if (argc != 3) {
		fprintf(stderr, "usage: bench prec iter\n");
		exit(1);
	}
	prec = strtoul(argv[1], NULL, 10);
	iter = strtol(argv[2], NULL, 10);

	mpd_init(&ctx, prec);
	/* no more MPD_MINALLOC changes after here */

	x0 = mpd_new(&ctx);
	y0 = mpd_new(&ctx);

	mpd_set_string(x0, "0.222", &ctx);
	mpd_set_string(y0, "0.333", &ctx);

	start_clock = clock();
	color_point(x0, y0, iter, &ctx);
	end_clock = clock();

	mpd_print(x0);
	fprintf(stderr, "time: %f\n\n", (double)(end_clock-start_clock)/(double)CLOCKS_PER_SEC);

	mpd_del(x0);
	mpd_del(y0);

	return 0;
}


